"""
 bbbLib.py

 Contains general functions, that have no script dependencies, so be used with other scripts.

 ChangeLog:
 22/10/06 - Created.
 15/11/06 - Change: fetchURL() - use safe_url
            Change: cleanHTML() - to use regex not find
            Change: tweeks to many other funcs. (eg decodeEntities())
            New: urlTextToASCII(), unicodeToAscii()
 06/12/06 - Change: DialogSelect()
 06/01/07 - Added searchRegEx(), findAllRegEx(), isImageURL(), isRSSLink(),
            isHTMLLink(), ViewTextFile(),listDir()


"""

__title__ = "bbbLib"
__version__ = "1.03"
__author__ = 'BigBellyBilly [BigBellyBilly@gmail.com]'
__date__ = '06-01-2007'

import sys,os.path
DIR_HOME= os.getcwd()
if DIR_HOME[-1]==';': DIR_HOME=DIR_HOME[0:-1]
if DIR_HOME[-1]!='\\': DIR_HOME=DIR_HOME+'\\'
DIR_SYSTEM = DIR_HOME + "system\\"
DIR_CACHE = DIR_HOME + "cache\\"
DIR_GFX = DIR_SYSTEM + "gfx\\"

import xbmc, xbmcgui
xbmc.output("loading module: " + __name__ + " Version: " + __version__ + " Date: " + __date__)
import os, re, unicodedata, traceback
import urllib
from threading import Thread
from string import strip, replace, find, rjust

try: Emulating = xbmcgui.Emulating
except: Emulating = False

# KEYPAD CODES
ACTION_UNKNOWN			= 0
ACTION_MOVE_LEFT 	    = 1 	# Dpad
ACTION_MOVE_RIGHT	    = 2		# Dpad
ACTION_MOVE_UP		    = 3		# Dpad
ACTION_MOVE_DOWN	    = 4		# Dpad
ACTION_PAGE_UP		    = 111	# trigger left
ACTION_PAGE_DOWN	    = 112	# trigger right
ACTION_SCROLL_UP		= 5		# trigger up
ACTION_SCROLL_DOWN		= 6		# trigger down
ACTION_HIGHLIGHT_ITEM	= 8
ACTION_PREVIOUS_MENU	= 10	# back btn
ACTION_SHOW_INFO	    = 117	# white button
ACTION_REMOTE_INFO		= 11	# info on remote
ACTION_PAUSE		    = 12	# remote
ACTION_STOP		        = 13	# remote
ACTION_NEXT_ITEM	    = 14	# remote Skip Next
ACTION_PREV_ITEM	    = 15	# remote Skip Previous
ACTION_SELECT_ITEM	    = 7     # A
ACTION_Y_BUTTON 	    = 34	# Y
ACTION_X_BUTTON 	    = 18    # X
ACTION_PARENT_DIR	    = 9     # B
ACTION_RECORD			= 2010  # record button on remote - assigned in keymap.xml

XBFONT_LEFT       = 0x00000000
XBFONT_RIGHT      = 0x00000001
XBFONT_CENTER_X   = 0x00000002
XBFONT_CENTER_Y   = 0x00000004
XBFONT_TRUNCATED  = 0x00000008

KBTYPE_ALPHA = -1
KBTYPE_NUMERIC = 0
KBTYPE_DATE = 1
KBTYPE_TIME = 2
KBTYPE_IP = 3

# GFX - for library funcs
BACKGROUND_FILENAME = DIR_GFX + 'background.png'
BACKG_BLANK_FILENAME = DIR_GFX + 'blank_background.png'
HEADER_FILENAME = DIR_GFX + 'header.png'
FOOTER_FILENAME = DIR_GFX + 'footer.png'
LOGO_FILENAME = DIR_GFX + 'logo.png'
PANEL_FILENAME = 'dialog-panel.png'
PANEL_FILENAME_MC360 = 'dialog-panel_mc360.png'
BTN_A_FILENAME = DIR_GFX + 'abutton.png'
BTN_B_FILENAME = DIR_GFX + 'bbutton.png'
BTN_X_FILENAME = DIR_GFX + 'xbutton.png'
BTN_Y_FILENAME = DIR_GFX + 'ybutton.png'
BTN_WHITE_FILENAME = DIR_GFX + 'whitebutton.png'
OVERSCAN_TL_FILENAME = DIR_GFX + 'overscan_tl.png'
OVERSCAN_BR_FILENAME = DIR_GFX + 'overscan_br.png'
LIST_DELETE_FILENAME = DIR_GFX + 'list_delete.png'
FRAME_FOCUS_FILENAME = DIR_GFX + 'frame_focus.png'
FRAME_NOFOCUS_FILENAME = DIR_GFX + 'frame_nofocus.png'
FRAME_NOFOCUS_LRG_FILENAME = DIR_GFX + 'frame_nofocus_large.png'
README_FILENAME = DIR_HOME + 'readme.txt'
CHANGELOG_FILENAME = DIR_HOME + 'changelog.txt'

dialog = xbmcgui.DialogProgress()
REZ_W = 720
REZ_H = 576

#######################################################################################################################    
# DEBUG - display indented information
#######################################################################################################################    
DEBUG = False
debugLvl = 0	# current indentation level
def debug( value ):
	global debugLvl
	if (DEBUG and value):
		try:
			if value[0] == ">": debugLvl += 2
			pad = rjust("", debugLvl)
			print( pad + str(value) )
			if value[0] == "<": debugLvl -= 2
		except:
			print "debug print failed"


#################################################################################################################
def messageOK(title, line1, line2='',line3=''):
	text = line1
	if line2: text += '\n' + line2
	if line3: text += '\n' + line3
	xbmcgui.Dialog().ok(title ,line1,line2,line3)

#################################################################################################################
def handleException(txt=''):
	try:
		title = "EXCEPTION: " + txt
		e=sys.exc_info()
		list = traceback.format_exception(e[0],e[1],e[2],3)
		text = ''
		for l in list:
			text += l
		messageOK(title, text)
	except: pass
#	sys.exit(1)

#################################################################################################################
# delete a single file
def deleteFile(filename):
	try:
		if fileExist(filename):
			debug("removing file: " + filename)
			os.remove(filename)
	except:
		pass


#################################################################################################################
def readFile(filename):
	doc = ""
	try:
		doc = file(filename).read()
	except:
		messageOK("Error","Unable to read file: " + filename)
		handleException("readFile()")
	return doc

#################################################################################################################
def fileExist(filename):
	exist = False
	try:
		if os.path.isfile(filename) and os.path.getsize(filename) > 0:
			exist = True
	except:
		pass
	return exist

#################################################################################################################
def gethostipbyname(host):
	ip = host
	if not isIP(host):
		try:
			hostEnt = nmb.NetBIOS().gethostbyname( host )
			ip = hostEnt[0].get_ip()
		except:
			ip = ''
	debug("gethostipbyname() ip=" + ip)
	return ip

#################################################################################################################
def isIP(host):
	return re.match('^\d+\.\d+\.\d+\.\d+$', host)

##############################################################################################
def doKeyboard(currentValue='', heading='', kbType=KBTYPE_ALPHA):
	if currentValue == None:
		currentValue = ''
	value = currentValue
	if kbType == KBTYPE_ALPHA:
		keyboard = xbmc.Keyboard(currentValue, heading)
		keyboard.doModal()
		if keyboard.isConfirmed:
			value = keyboard.getText().strip()
		else:
			value = None
	else:
		d = xbmcgui.Dialog()
		value = d.numeric(kbType, heading)

	return value

##############################################################################################
def setResolution(win):
	debug("> setResolution()")
	try:
		# '1080i'=0, '720p'=1, '480p'=2, '480p16x9'=3, 'ntsc'=4, 'ntsc16x9'=5, 'pal'=6, 'pal16x9'=7, 'pal60'=8, 'pal6016x9'=9}
		rez = 6		# PAL
		currRez = win.getResolution()
		debug("currRez="+str(currRez))
		if (currRez != rez) and rez > 1:
			# If resolutions differ calculate widescreen offset
			# check if current resolution is 16x9
			offset = 0
			if (currRez == 0 or currRez % 2): iCur16x9 = 1
			else: iCur16x9 = 0
			debug("iCur16x9=" + str(iCur16x9))

			# check if skinned resolution is 16x9
			if (rez == 0 or rez % 2): i16x9 = 1
			else: i16x9 = 0
			debug("i16x9=" + str(i16x9))

			# calculate offset
			offset = iCur16x9 - i16x9
			rez += offset
	except:
		handleException()

	win.setCoordinateResolution(rez)
	debug("< setResolution() final rez mode="+str(rez))


#################################################################################################################
# Dialog with options
#################################################################################################################
class DialogSelect(xbmcgui.WindowDialog):
	def __init__(self):
		if Emulating: xbmcgui.WindowDialog.__init__(self)

		setResolution(self)

		self.selectedPos = -1		# EXIT option
		self.panelCI = None
		self.menuCL = None
		self.lastAction = 0

	def setup(self, title='Menu',width=430, height=450, xpos=-1, ypos=-1, textColor='0xFFFFFFFF', \
			  imageWidth=0,imageHeight=20, itemHeight=20, rows=0,
			  useY=False, useX=False, isDelete=False, \
			  panel=PANEL_FILENAME):
		debug("> DialogSelect().setup() useY="+str(useY) + " useX=" + str(useX) + " isDelete=" + str(isDelete))

		self.useY = useY
		self.useX = useX
		self.isDelete = isDelete
		offsetLabelX = 15
		offsetListX = 20
		listW = width - offsetListX - 5
		if imageHeight > itemHeight:
			itemHeight = imageHeight
		titleH=30

		if title:
			offsetLabelY = 15
			offsetListY = offsetLabelY + titleH
		else:
			offsetLabelY = 0
			offsetListY = 5

		# calc height based on rows, otherwise use supplied height
		if rows > 0:
			rows += 3		 # add 'exit' + extra rows
			listH = (rows * itemHeight)
		else:
			listH = height

		if height > listH:
			height = listH			# shrink height to fit list
		elif height < listH:
			listH = height-itemHeight			# shrink list to height

		height += offsetListY		# add space onto for title + gap

		# center on screen
		if xpos < 0:
			xpos = int((REZ_W /2) - (width /2))
		if ypos < 0:
			ypos = int((REZ_H /2) - (height /2)) + 10

		try:
			self.removeControl(self.panelCI)
		except:
			pass
		try:
			self.panelCI = xbmcgui.ControlImage(xpos, ypos, width, height, panel)
			self.addControl(self.panelCI)
		except:
			pass

		if isDelete:
			title += " (Delete: Highlight & Press A)"
		if title:
			self.titleCL = xbmcgui.ControlLabel(xpos+offsetLabelX, ypos+offsetLabelY, listW, titleH, \
												title, 'font13', '0xFFFFFF00', alignment=XBFONT_CENTER_X)
			self.addControl(self.titleCL)

		try:
			self.removeControl(self.menuCL)
		except:
			pass
		if not self.isDelete:
			self.menuCL = xbmcgui.ControlList(xpos+offsetListX, ypos+offsetListY, \
											listW, listH, textColor=textColor, \
											imageWidth=imageWidth, imageHeight=imageHeight, \
											itemHeight=itemHeight, itemTextXOffset=0, space=0)
		else:
			self.menuCL = xbmcgui.ControlList(xpos+offsetListX, ypos+offsetListY, \
											listW, listH, textColor=textColor, \
											imageWidth=imageWidth, imageHeight=imageHeight, \
											itemHeight=itemHeight, buttonFocusTexture=LIST_DELETE_FILENAME, \
											itemTextXOffset=0, space=0)
		self.addControl(self.menuCL)
		self.menuCL.setPageControlVisible(False)
		debug("< DialogSelect().setup()")

	def onAction(self, action):
		debug("DialogSelect.onAction()")
		if action == 0:
			return
		self.lastAction = action
		if action == ACTION_PREVIOUS_MENU or action == ACTION_PARENT_DIR:
			self.selectedPos = -1
			self.close()
		elif (self.useY and action == ACTION_Y_BUTTON) or \
			 (self.useX and action == ACTION_X_BUTTON):
			self.selectedPos = self.menuCL.getSelectedPosition()
			self.close()

	def onControl(self, control):
		debug("DialogSelect.onControl()")
		try:
			self.selectedPos = self.menuCL.getSelectedPosition()
		except: pass
		self.close()

	def setMenu(self, menu, icons=[]):
		debug("> DialogSelect.setMenu()")
		self.menuCL.reset()
		self.menuCL.addItem("Exit")
		if menu:
			for idx in range(len(menu)):
				opt = menu[idx]
				try:
					img = icons[idx]
				except:
					img = ''
				if isinstance(opt, xbmcgui.ListItem):
					if img:
						opt.setIconImage(img)
					self.menuCL.addItem(opt)
				else:
					self.menuCL.addItem(xbmcgui.ListItem(opt.strip(), '', thumbnailImage=img))

			if self.selectedPos > len(menu):
				self.selectedPos = len(menu)-1
			self.menuCL.selectItem(self.selectedPos)
		debug("< DialogSelect.setMenu()")

	# show this dialog and wait until it's closed
	def ask(self, menu, selectIdx=-1, icons=[]):
		debug ("> DialogSelect().ask() selectIdx=" +str(selectIdx))

		self.selectedPos = selectIdx+1	# allow for exit option
		self.setMenu(menu, icons)
		self.setFocus(self.menuCL)
		self.doModal()
		self.selectedPos -= 1
		debug ("< DialogSelect.ask() selectedPos: " + str(self.selectedPos))
		return self.selectedPos, self.lastAction

#######################################################################################################################    
def parseDocList(html, startStr, endStr, regex):
	debug("> parseDocList() startStr='"+startStr+"' endStr='"+endStr+"'")

	# extract section
	startPos = 0
	matchList = []
	if startStr:
		startPos = find(html,startStr)
		if startPos == -1:	# not found
			debug("start not found")

	if startPos >= 0:
		if endStr:
			endPos = find(html, endStr, startPos+1)
			if endPos == -1: # not found
				debug("end not found")
				partHTML = html[startPos:]
			else:
				partHTML = html[startPos:endPos]
		else:
			partHTML = html[startPos:]

		findRe = re.compile(regex, re.DOTALL + re.MULTILINE + re.IGNORECASE)
		matchList = findRe.findall(partHTML)

	debug("< parseDocList() sz= " + str(len(matchList)))
	return matchList

#################################################################################################################
# look for html between < and > chars and remove it
def cleanHTML(data):
	reobj = re.compile('<.+?>', re.IGNORECASE+re.DOTALL+re.MULTILINE)
	return (re.sub(reobj, '', data)).strip()

#################################################################################################################
def HTTPErrorCode(code):
	debug("HTTPErrorCode: " + str(code))
	if code == 301:
		title = 'HTTPError 301'
		txt = 'Page Moved Permanently'
	elif code == 302:
		title = 'HTTPError 302'
		txt = 'Page Moved Temporarily'
	elif code == 404:
		title = 'HTTPError 404'
		txt = 'Page Not Found.','Please try later.'
	elif code == 502:
		title = 'HTTPError 502'
		txt = 'Service temporarily overloaded.','Please try later.'
	elif code == 500:
		title = 'HTTPError 500'
		txt = 'Internal Server Error, Has your account expired?'
	else:
		title = 'Error: ' + str(code)
		txt = ''
	return title, txt

#################################################################################################################
# Class to return a char or string pixel length. Thanks to madtw.
# Until truncation of labels on ControlButtons is working.
#################################################################################################################
class FontAttr:
	def __init__(self):
		self.adjust = {'font11':-3, 'font12':-2, 'font13':-1, 'font14':0, 'font18':4}

		# base widths based on font14
		self.width = [ \
			0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
			0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
			0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
			#       SP   !   "   #   $   %   &   '
			0,  0,  6,  6,  8, 14, 10, 19, 13,  4,
			#(   )   *   +   ,   -   .   /   0   1
			7,  7, 24, 14,  6,  7,  6,  7, 10, 10,
			#2   3   4   5   6   7   8   9   :  ;
			10, 10, 10, 10, 10, 10, 10, 10,  7,  7,
			#<   =   >   ?   @   A   B   C   D   E
			14, 14, 14,  9, 17, 11, 11, 11, 13, 11,
			#F   G   H   I   J   K   L   M   N   O
			10, 13, 13,  7,  8, 11,  9, 15, 13, 13,
			#P   Q   R   S   T   U   V   W   X   Y
			10, 13, 12, 11, 11, 12, 11, 17, 11, 11,
			#Z   [   \   ]   ^   _   `   a   b   c
			11,  7,  7,  7, 14, 10, 10, 10, 11,  9,
			#d   e   f   g   h   i   j   k   l   m
			11, 10,  6, 11, 11,  4,  5,  9,  4, 16,
			#n   o   p   q   r   s   t   u   v   w
			11, 10, 11, 11,  7,  8,  6, 11,  9, 14,
			#x   y   z   {   |   }   ~
			9,  9,  8,  9,  7,  9, 14
		]

	def getWidth( self, c, font = 'font14' ):
		try:
			w = self.width[ord(c)]
			adj = self.adjust[font]
			w += adj
		except:
			w = 14			# unknown ch, use avg width
		return w

	def getTextWidth(self, txt, font = 'font14'):
		len = 0
		for c in txt:
			len += self.getWidth(c, font)
		return len

	def truncate(self, maxWidth, text, font = 'font14'):
		sz = len(text)
		newText = ''
		for i in range(sz, 0, -1):
			newText = text[0:i]
			newWidth = self.getTextWidth(newText, font)
			if newWidth < maxWidth:
				break

		return newText

#######################################################################################################
class AdjustOverscan(xbmcgui.Window):
	def __init__(self):
		if Emulating: xbmcgui.Window.__init__(self)
		debug("> AdjustOverscan() init()")

		setResolution(self)
		self.isTL = True
		self.backg = xbmcgui.ControlImage(0,0, REZ_W, REZ_H, BACKG_BLANK_FILENAME)
		self.info = xbmcgui.ControlTextBox(150, 150, 400, 300, 'font12')
		self.tlCB = xbmcgui.ControlButton(0, 0, 0, 0, '', focusTexture=OVERSCAN_TL_FILENAME)
		self.brCB = xbmcgui.ControlButton(0, 0, 0, 0, '', focusTexture=OVERSCAN_BR_FILENAME)

		debug("< AdjustOverscan() init()")

	###################################################################################################
	def onAction(self, action):
		if action == ACTION_PREVIOUS_MENU:
			self.close()
		elif action == ACTION_SELECT_ITEM:
			self.isTL = not self.isTL
			self.draw()
		elif action == ACTION_MOVE_LEFT or action == ACTION_MOVE_RIGHT or \
			 action == ACTION_MOVE_UP or action == ACTION_MOVE_DOWN:
			if action == ACTION_MOVE_LEFT:
				if self.isTL:
					self.overscanL -= 1
				else:
					self.overscanR += 1
			elif action == ACTION_MOVE_RIGHT:
				if self.isTL:
					self.overscanL += 1
				else:
					self.overscanR -= 1
			elif action == ACTION_MOVE_UP:
				if self.isTL:
					self.overscanT -= 1
				else:
					self.overscanB += 1
			elif action == ACTION_MOVE_DOWN:
				if self.isTL:
					self.overscanT += 1
				else:
					self.overscanB -= 1
			self.draw()

	###################################################################################################
	def onControl(self, control):
		pass

	##############################################################################################
	def draw(self):
		xbmcgui.lock()
		text = 'To Adjust Overscan:\n\nDPAD = Move corner.  A Button = Change corner.  BACK = Exit.\n\nAdjusting: '
		if self.isTL:
			text +=  'TOP LEFT'
		else:
			text += 'BOTTOM RIGHT'
		text += '\n\nOverscan TOP: ' + str(self.overscanT) + \
			   '\nOverscan LEFT: ' + str(self.overscanL) + \
			   '\nOverscan BOTTOM: ' + str(self.overscanB) + \
			   '\nOverscan RIGHT: ' + str(self.overscanR)
		self.info.setText(text)

		if self.isTL:
			self.tlCB.setPosition(self.overscanL, self.overscanT)
			self.brCB.setLabel('')
			self.tlCB.setLabel('Adjust')
			self.setFocus(self.tlCB)
		else:
			x = REZ_W - self.overscanR - 57	# img w
			y = REZ_H - self.overscanB - 57 # img h
			self.brCB.setPosition(x, y)
			self.tlCB.setLabel('')
			self.brCB.setLabel('Adjust')
			self.setFocus(self.brCB)

		xbmcgui.unlock()

	##############################################################################################
	def ask(self, top=30, left=30, bottom=30, right=30):
		debug("> ask()")
		# OVERSCAN
		self.overscanT = top
		self.overscanL = left
		self.overscanB = bottom
		self.overscanR = right
		self.addControl(self.backg)
		self.addControl(self.info)
		self.addControl(self.tlCB)
		self.addControl(self.brCB)
		self.draw()
		self.doModal()
		debug("< ask()")
		return self.overscanT, self.overscanL, self.overscanB, self.overscanR

#################################################################################################################
# Convert a text string containing &#x<hex_value> to ascii ch
#################################################################################################################
def urlTextToASCII(text):
	try:
		compile_obj = re.compile('(&#x(.*?);)',  re.IGNORECASE + re.MULTILINE + re.DOTALL)
		match_obj = compile_obj.findall(text)
		for match in match_obj:
			ch = chr(int('0x'+match[1], 16))
			text = text.replace(match[0], ch)

		compile_obj = re.compile('(&#(\d+);)',  re.IGNORECASE + re.MULTILINE + re.DOTALL)
		match_obj = compile_obj.findall(text)
		for match in match_obj:
			ch = chr(int(match[1]))
			text = text.replace(match[0], ch)
		debug("urlTextToASCII() DONE")
	except:
		debug("urlTextToASCII() exception")
	return text


#################################################################################################################
# Thanks to Arboc for this
#################################################################################################################
def unicodeToAscii(txt, charset='utf8'):
	try:
		newtxt = txt.decode(charset)
		newtxt = unicodedata.normalize('NFKD', newtxt).encode('ASCII','replace')
		return newtxt
	except:
		return txt

#################################################################################################################
# Thanks to Arboc for contributing most of the translation in this function. 
#################################################################################################################
def decodeEntities(txt):
	txt = txt.replace('\n','')
	txt = txt.replace('\r','')
	txt = txt.replace('\t','')

# % values
	if find(txt,'%') >= 0:
		txt = txt.replace('%21', "!")
		txt = txt.replace('%22', '"')
		txt = txt.replace('%25', "%")
		txt = txt.replace('%26', "&")
		txt = txt.replace('%27', "'")
		txt = txt.replace('%28', "(")
		txt = txt.replace('%29', ")")
		txt = txt.replace('%2a', "*")
		txt = txt.replace('%2b', "+")
		txt = txt.replace('%2c', ",")
		txt = txt.replace('%2d', "-")
		txt = txt.replace('%2e', ".")
		txt = txt.replace('%3a', ":")
		txt = txt.replace('%3b', ";")
		txt = txt.replace('%3f', "?")
		txt = txt.replace('%40', "@")

# 
	if find(txt,"&#") >= 0:
		txt = txt.replace('&#034;','"')
		txt = txt.replace('&#039;','\'')
		txt = txt.replace('&#13;&#10;','\n')
		txt = txt.replace('&#146;', "'")
		txt = txt.replace('&#156;', "oe")
		txt = txt.replace('&#160;', " ") # no-break space = non-breaking space U+00A0 ISOnum
		txt = txt.replace('&#161;', "!") # inverted exclamation mark U+00A1 ISOnum
		txt = txt.replace('&#162;', "c") # cent sign U+00A2 ISOnum
		txt = txt.replace('&#163;', "p") # pound sign U+00A3 ISOnum
		txt = txt.replace('&#164;', "$") # currency sign U+00A4 ISOnum
		txt = txt.replace('&#165;', "y") # yen sign = yuan sign U+00A5 ISOnum
		txt = txt.replace('&#166;', "|") # broken bar = broken vertical bar U+00A6 ISOnum
		txt = txt.replace('&#167;', "S") # section sign U+00A7 ISOnum
		txt = txt.replace('&#168;', "''") # diaeresis = spacing diaeresis U+00A8 ISOdia
		txt = txt.replace('&#169;', "(c)") # copyright sign U+00A9 ISOnum
		txt = txt.replace('&#170;', "e") # feminine ordinal indicator U+00AA ISOnum
		txt = txt.replace('&#171;', '"') # left-pointing double angle quotation mark = left pointing guillemet U+00AB ISOnum
		txt = txt.replace('&#172;', "-.") # not sign U+00AC ISOnum
		txt = txt.replace('&#173;', "-") # soft hyphen = discretionary hyphen U+00AD ISOnum
		txt = txt.replace('&#174;', "(R)") # registered sign = registered trade mark sign U+00AE ISOnum
		txt = txt.replace('&#175;', "-") # macron = spacing macron = overline = APL overbar U+00AF ISOdia
		txt = txt.replace('&#176;', "o") # degree sign U+00B0 ISOnum
		txt = txt.replace('&#177;', "+-") # plus-minus sign = plus-or-minus sign U+00B1 ISOnum
		txt = txt.replace('&#178;', "2") # superscript two = superscript digit two = squared U+00B2 ISOnum
		txt = txt.replace('&#179;', "3") # superscript three = superscript digit three = cubed U+00B3 ISOnum
		txt = txt.replace('&#180;', " ") # acute accent = spacing acute U+00B4 ISOdia
		txt = txt.replace('&#181;', "u") # micro sign U+00B5 ISOnum
		txt = txt.replace('&#182;', "|p") # pilcrow sign = paragraph sign U+00B6 ISOnum
		txt = txt.replace('&#183;', ".") # middle dot = Georgian comma = Greek middle dot U+00B7 ISOnum
		txt = txt.replace('&#184;', " ") # cedilla = spacing cedilla U+00B8 ISOdia
		txt = txt.replace('&#185;', "1") # superscript one = superscript digit one U+00B9 ISOnum
		txt = txt.replace('&#186;', "o") # masculine ordinal indicator U+00BA ISOnum
		txt = txt.replace('&#187;', '"') # right-pointing double angle quotation mark = right pointing guillemet U+00BB ISOnum
		txt = txt.replace('&#188;', "1/4") # vulgar fraction one quarter = fraction one quarter U+00BC ISOnum
		txt = txt.replace('&#189;', "1/2") # vulgar fraction one half = fraction one half U+00BD ISOnum
		txt = txt.replace('&#190;', "3/4") # vulgar fraction three quarters = fraction three quarters U+00BE ISOnum
		txt = txt.replace('&#191;', "?") # inverted question mark = turned question mark U+00BF ISOnum
		txt = txt.replace('&#192;', "A") # latin capital letter A with grave = latin capital letter A grave U+00C0 ISOlat1
		txt = txt.replace('&#193;', "A") # latin capital letter A with acute U+00C1 ISOlat1
		txt = txt.replace('&#194;', "A") # latin capital letter A with circumflex U+00C2 ISOlat1
		txt = txt.replace('&#195;', "A") # latin capital letter A with tilde U+00C3 ISOlat1
		txt = txt.replace('&#196;', "A") # latin capital letter A with diaeresis U+00C4 ISOlat1
		txt = txt.replace('&#197;', "A") # latin capital letter A with ring above = latin capital letter A ring U+00C5 ISOlat1
		txt = txt.replace('&#198;', "AE") # latin capital letter AE = latin capital ligature AE U+00C6 ISOlat1
		txt = txt.replace('&#199;', "C") # latin capital letter C with cedilla U+00C7 ISOlat1
		txt = txt.replace('&#200;', "E") # latin capital letter E with grave U+00C8 ISOlat1
		txt = txt.replace('&#201;', "E") # latin capital letter E with acute U+00C9 ISOlat1
		txt = txt.replace('&#202;', "E") # latin capital letter E with circumflex U+00CA ISOlat1
		txt = txt.replace('&#203;', "E") # latin capital letter E with diaeresis U+00CB ISOlat1
		txt = txt.replace('&#204;', "I") # latin capital letter I with grave U+00CC ISOlat1
		txt = txt.replace('&#205;', "I") # latin capital letter I with acute U+00CD ISOlat1
		txt = txt.replace('&#206;', "I") # latin capital letter I with circumflex U+00CE ISOlat1
		txt = txt.replace('&#207;', "I") # latin capital letter I with diaeresis U+00CF ISOlat1
		txt = txt.replace('&#208;', "D") # latin capital letter ETH U+00D0 ISOlat1
		txt = txt.replace('&#209;', "N") # latin capital letter N with tilde U+00D1 ISOlat1
		txt = txt.replace('&#210;', "O") # latin capital letter O with grave U+00D2 ISOlat1
		txt = txt.replace('&#211;', "O") # latin capital letter O with acute U+00D3 ISOlat1
		txt = txt.replace('&#212;', "O") # latin capital letter O with circumflex U+00D4 ISOlat1
		txt = txt.replace('&#213;', "O") # latin capital letter O with tilde U+00D5 ISOlat1
		txt = txt.replace('&#214;', "O") # latin capital letter O with diaeresis U+00D6 ISOlat1
		txt = txt.replace('&#215;', "x") # multiplication sign U+00D7 ISOnum
		txt = txt.replace('&#216;', "O") # latin capital letter O with stroke = latin capital letter O slash U+00D8 ISOlat1
		txt = txt.replace('&#217;', "U") # latin capital letter U with grave U+00D9 ISOlat1
		txt = txt.replace('&#218;', "U") # latin capital letter U with acute U+00DA ISOlat1
		txt = txt.replace('&#219;', "U") # latin capital letter U with circumflex U+00DB ISOlat1
		txt = txt.replace('&#220;', "U") # latin capital letter U with diaeresis U+00DC ISOlat1
		txt = txt.replace('&#221;', "Y") # latin capital letter Y with acute U+00DD ISOlat1
		txt = txt.replace('&#222;', "D") # latin capital letter THORN U+00DE ISOlat1
		txt = txt.replace('&#223;', "SS") # latin small letter sharp s = ess-zed U+00DF ISOlat1
		txt = txt.replace('&#224;', "a") # latin small letter a with grave = latin small letter a grave U+00E0 ISOlat1
		txt = txt.replace('&#225;', "a") # latin small letter a with acute U+00E1 ISOlat1
		txt = txt.replace('&#226;', "a") # latin small letter a with circumflex U+00E2 ISOlat1
		txt = txt.replace('&#227;', "a") # latin small letter a with tilde U+00E3 ISOlat1
		txt = txt.replace('&#228;', "a") # latin small letter a with diaeresis U+00E4 ISOlat1
		txt = txt.replace('&#229;', "a") # latin small letter a with ring above = latin small letter a ring U+00E5 ISOlat1
		txt = txt.replace('&#230;', "ae") # latin small letter ae = latin small ligature ae U+00E6 ISOlat1
		txt = txt.replace('&#231;', "c") # latin small letter c with cedilla U+00E7 ISOlat1
		txt = txt.replace('&#232;', "e") # latin small letter e with grave U+00E8 ISOlat1
		txt = txt.replace('&#233;', "e") # latin small letter e with acute U+00E9 ISOlat1
		txt = txt.replace('&#234;', "e") # latin small letter e with circumflex U+00EA ISOlat1
		txt = txt.replace('&#235;', "e") # latin small letter e with diaeresis U+00EB ISOlat1
		txt = txt.replace('&#236;', "i") # latin small letter i with grave U+00EC ISOlat1
		txt = txt.replace('&#237;', "i") # latin small letter i with acute U+00ED ISOlat1
		txt = txt.replace('&#238;', "i") # latin small letter i with circumflex U+00EE ISOlat1
		txt = txt.replace('&#239;', "i") # latin small letter i with diaeresis U+00EF ISOlat1
		txt = txt.replace('&#240;', "d") # latin small letter eth U+00F0 ISOlat1
		txt = txt.replace('&#241;', "n") # latin small letter n with tilde U+00F1 ISOlat1
		txt = txt.replace('&#242;', "o") # latin small letter o with grave U+00F2 ISOlat1
		txt = txt.replace('&#243;', "o") # latin small letter o with acute U+00F3 ISOlat1
		txt = txt.replace('&#244;', "o") # latin small letter o with circumflex U+00F4 ISOlat1
		txt = txt.replace('&#245;', "o") # latin small letter o with tilde U+00F5 ISOlat1
		txt = txt.replace('&#246;', "o") # latin small letter o with diaeresis U+00F6 ISOlat1
		txt = txt.replace('&#247;', "/") # division sign U+00F7 ISOnum
		txt = txt.replace('&#248;', "o") # latin small letter o with stroke = latin small letter o slash U+00F8 ISOlat1
		txt = txt.replace('&#249;', "u") # latin small letter u with grave U+00F9 ISOlat1
		txt = txt.replace('&#250;', "u") # latin small letter u with acute U+00FA ISOlat1
		txt = txt.replace('&#251;', "u") # latin small letter u with circumflex U+00FB ISOlat1
		txt = txt.replace('&#252;', "u") # latin small letter u with diaeresis U+00FC ISOlat1
		txt = txt.replace('&#253;', "y") # latin small letter y with acute U+00FD ISOlat1
		txt = txt.replace('&#254;', "d") # latin small letter thorn U+00FE ISOlat1
		txt = txt.replace('&#255;', "y") # latin small letter y with diaeresis U+00FF ISOlat1
		txt = txt.replace('&#338;', "OE") # latin capital ligature OE U+0152 ISOlat2
		txt = txt.replace('&#339;', "oe") # latin small ligature oe U+0153 ISOlat2
		txt = txt.replace('&#34;', '"') # quotation mark = APL quote U+0022 ISOnum
		txt = txt.replace('&#352;', "S") # latin capital letter S with caron U+0160 ISOlat2
		txt = txt.replace('&#353;', "s") # latin small letter s with caron U+0161 ISOlat2
		txt = txt.replace('&#376;', "Y") # latin capital letter Y with diaeresis U+0178 ISOlat2
		txt = txt.replace('&#38;', "&") # ampersand U+0026 ISOnum
		txt = txt.replace('&#39;','\'')
		txt = txt.replace('&#402;', "f") # latin small f with hook = function = florin U+0192 ISOtech
		txt = txt.replace('&#60;', "<") # less-than sign U+003C ISOnum
		txt = txt.replace('&#62;', ">") # greater-than sign U+003E ISOnum
		txt = txt.replace('&#710;', "") # modifier letter circumflex accent U+02C6 ISOpub
		txt = txt.replace('&#732;', "~") # small tilde U+02DC ISOdia
		txt = txt.replace('&#8194;', "") # en space U+2002 ISOpub
		txt = txt.replace('&#8195;', " ") # em space U+2003 ISOpub
		txt = txt.replace('&#8201;', " ") # thin space U+2009 ISOpub
		txt = txt.replace('&#8204;', "|") # zero width non-joiner U+200C NEW RFC 2070
		txt = txt.replace('&#8205;', "|") # zero width joiner U+200D NEW RFC 2070
		txt = txt.replace('&#8206;', "") # left-to-right mark U+200E NEW RFC 2070
		txt = txt.replace('&#8207;', "") # right-to-left mark U+200F NEW RFC 2070
		txt = txt.replace('&#8211;', "--") # en dash U+2013 ISOpub
		txt = txt.replace('&#8212;', "---") # em dash U+2014 ISOpub
		txt = txt.replace('&#8216;', '"') # left single quotation mark U+2018 ISOnum
		txt = txt.replace('&#8217;', '"') # right single quotation mark U+2019 ISOnum
		txt = txt.replace('&#8218;', '"') # single low-9 quotation mark U+201A NEW
		txt = txt.replace('&#8220;', '"') # left double quotation mark U+201C ISOnum
		txt = txt.replace('&#8221;', '"') # right double quotation mark U+201D ISOnum
		txt = txt.replace('&#8222;', '"') # double low-9 quotation mark U+201E NEW
		txt = txt.replace('&#8224;', "|") # dagger U+2020 ISOpub
		txt = txt.replace('&#8225;', "||") # double dagger U+2021 ISOpub
		txt = txt.replace('&#8226;', "o") # bullet = black small circle U+2022 ISOpub
		txt = txt.replace('&#8230;', "...") # horizontal ellipsis = three dot leader U+2026 ISOpub
		txt = txt.replace('&#8240;', "%0") # per mille sign U+2030 ISOtech
		txt = txt.replace('&#8242;', "'") # prime = minutes = feet U+2032 ISOtech
		txt = txt.replace('&#8243;', "''") # double prime = seconds = inches U+2033 ISOtech
		txt = txt.replace('&#8249;', '"') # single left-pointing angle quotation mark U+2039 ISO proposed
		txt = txt.replace('&#8250;', '"') # single right-pointing angle quotation mark U+203A ISO proposed
		txt = txt.replace('&#8254;', "-") # overline = spacing overscore U+203E NEW
		txt = txt.replace('&#8260;', "/") # fraction slash U+2044 NEW
		txt = txt.replace('&#8364;', "EU$") # euro sign U+20AC NEW
		txt = txt.replace('&#8465;', "|I") # blackletter capital I = imaginary part U+2111 ISOamso
		txt = txt.replace('&#8472;', "|P") # script capital P = power set = Weierstrass p U+2118 ISOamso
		txt = txt.replace('&#8476;', "|R") # blackletter capital R = real part symbol U+211C ISOamso
		txt = txt.replace('&#8482;', "tm") # trade mark sign U+2122 ISOnum
		txt = txt.replace('&#8501;', "%") # alef symbol = first transfinite cardinal U+2135 NEW
		txt = txt.replace('&#8592;', "<-") # leftwards arrow U+2190 ISOnum
		txt = txt.replace('&#8593;', "^") # upwards arrow U+2191 ISOnum
		txt = txt.replace('&#8594;', "->") # rightwards arrow U+2192 ISOnum
		txt = txt.replace('&#8595;', "v") # downwards arrow U+2193 ISOnum
		txt = txt.replace('&#8596;', "<->") # left right arrow U+2194 ISOamsa
		txt = txt.replace('&#8629;', "<-'") # downwards arrow with corner leftwards = carriage return U+21B5 NEW
		txt = txt.replace('&#8656;', "<=") # leftwards double arrow U+21D0 ISOtech
		txt = txt.replace('&#8657;', "^") # upwards double arrow U+21D1 ISOamsa
		txt = txt.replace('&#8658;', "=>") # rightwards double arrow U+21D2 ISOtech
		txt = txt.replace('&#8659;', "v") # downwards double arrow U+21D3 ISOamsa
		txt = txt.replace('&#8660;', "<=>") # left right double arrow U+21D4 ISOamsa
		txt = txt.replace('&#8764;', "~") # tilde operator = varies with = similar to U+223C ISOtech
		txt = txt.replace('&#8773;', "~=") # approximately equal to U+2245 ISOtech
		txt = txt.replace('&#8776;', "~~") # almost equal to = asymptotic to U+2248 ISOamsr
		txt = txt.replace('&#8800;', "!=") # not equal to U+2260 ISOtech
		txt = txt.replace('&#8801;', "==") # identical to U+2261 ISOtech
		txt = txt.replace('&#8804;', "<=") # less-than or equal to U+2264 ISOtech
		txt = txt.replace('&#8805;', ">=") # greater-than or equal to U+2265 ISOtech
		txt = txt.replace('&#8901;', ".") # dot operator U+22C5 ISOamsb
		txt = txt.replace('&#913;', "Alpha") # greek capital letter alpha U+0391
		txt = txt.replace('&#914;', "Beta") # greek capital letter beta U+0392
		txt = txt.replace('&#915;', "Gamma") # greek capital letter gamma U+0393 ISOgrk3
		txt = txt.replace('&#916;', "Delta") # greek capital letter delta U+0394 ISOgrk3
		txt = txt.replace('&#917;', "Epsilon") # greek capital letter epsilon U+0395
		txt = txt.replace('&#918;', "Zeta") # greek capital letter zeta U+0396
		txt = txt.replace('&#919;', "Eta") # greek capital letter eta U+0397
		txt = txt.replace('&#920;', "Theta") # greek capital letter theta U+0398 ISOgrk3
		txt = txt.replace('&#921;', "Iota") # greek capital letter iota U+0399
		txt = txt.replace('&#922;', "Kappa") # greek capital letter kappa U+039A
		txt = txt.replace('&#923;', "Lambda") # greek capital letter lambda U+039B ISOgrk3
		txt = txt.replace('&#924;', "Mu") # greek capital letter mu U+039C
		txt = txt.replace('&#925;', "Nu") # greek capital letter nu U+039D
		txt = txt.replace('&#926;', "Xi") # greek capital letter xi U+039E ISOgrk3
		txt = txt.replace('&#927;', "Omicron") # greek capital letter omicron U+039F
		txt = txt.replace('&#928;', "Pi") # greek capital letter pi U+03A0 ISOgrk3
		txt = txt.replace('&#929;', "Rho") # greek capital letter rho U+03A1
		txt = txt.replace('&#931;', "Sigma") # greek capital letter sigma U+03A3 ISOgrk3
		txt = txt.replace('&#932;', "Tau") # greek capital letter tau U+03A4
		txt = txt.replace('&#933;', "Upsilon") # greek capital letter upsilon U+03A5 ISOgrk3
		txt = txt.replace('&#934;', "Phi") # greek capital letter phi U+03A6 ISOgrk3
		txt = txt.replace('&#935;', "Chi") # greek capital letter chi U+03A7
		txt = txt.replace('&#936;', "Psi") # greek capital letter psi U+03A8 ISOgrk3
		txt = txt.replace('&#937;', "Omega") # greek capital letter omega U+03A9 ISOgrk3
		txt = txt.replace('&#945;', "alpha") # greek small letter alpha U+03B1 ISOgrk3
		txt = txt.replace('&#946;', "beta") # greek small letter beta U+03B2 ISOgrk3
		txt = txt.replace('&#947;', "gamma") # greek small letter gamma U+03B3 ISOgrk3
		txt = txt.replace('&#948;', "delta") # greek small letter delta U+03B4 ISOgrk3
		txt = txt.replace('&#949;', "epsilon") # greek small letter epsilon U+03B5 ISOgrk3
		txt = txt.replace('&#950;', "zeta") # greek small letter zeta U+03B6 ISOgrk3
		txt = txt.replace('&#951;', "eta") # greek small letter eta U+03B7 ISOgrk3
		txt = txt.replace('&#952;', "theta") # greek small letter theta U+03B8 ISOgrk3
		txt = txt.replace('&#953;', "iota") # greek small letter iota U+03B9 ISOgrk3
		txt = txt.replace('&#954;', "kappa") # greek small letter kappa U+03BA ISOgrk3
		txt = txt.replace('&#955;', "lambda") # greek small letter lambda U+03BB ISOgrk3
		txt = txt.replace('&#956;', "mu") # greek small letter mu U+03BC ISOgrk3
		txt = txt.replace('&#957;', "nu") # greek small letter nu U+03BD ISOgrk3
		txt = txt.replace('&#958;', "xi") # greek small letter xi U+03BE ISOgrk3
		txt = txt.replace('&#959;', "omicron") # greek small letter omicron U+03BF NEW
		txt = txt.replace('&#960;', "pi") # greek small letter pi U+03C0 ISOgrk3
		txt = txt.replace('&#961;', "rho") # greek small letter rho U+03C1 ISOgrk3
		txt = txt.replace('&#962;', "sigma") # greek small letter final sigma U+03C2 ISOgrk3
		txt = txt.replace('&#963;', "sigma") # greek small letter sigma U+03C3 ISOgrk3
		txt = txt.replace('&#964;', "tau") # greek small letter tau U+03C4 ISOgrk3
		txt = txt.replace('&#965;', "upsilon") # greek small letter upsilon U+03C5 ISOgrk3
		txt = txt.replace('&#966;', "phi") # greek small letter phi U+03C6 ISOgrk3
		txt = txt.replace('&#967;', "chi") # greek small letter chi U+03C7 ISOgrk3
		txt = txt.replace('&#968;', "psi") # greek small letter psi U+03C8 ISOgrk3
		txt = txt.replace('&#969;', "omega") # greek small letter omega U+03C9 ISOgrk3
		txt = txt.replace('&#977;', "theta") # greek small letter theta symbol U+03D1 NEW
		txt = txt.replace('&#978;', "upsilon") # greek upsilon with hook symbol U+03D2 NEW
		txt = txt.replace('&#982;', "pi") # greek pi symbol U+03D6 ISOgrk3


	if find(txt,'&') >= 0:
		txt = txt.replace('&Aacute;', "A") # latin capital letter A with acute U+00C1 ISOlat1
		txt = txt.replace('&aacute;', "a") # latin small letter a with acute U+00E1 ISOlat1
		txt = txt.replace('&Acirc;', "A") # latin capital letter A with circumflex U+00C2 ISOlat1
		txt = txt.replace('&acirc;', "a") # latin small letter a with circumflex U+00E2 ISOlat1
		txt = txt.replace('&acute;', " ") # acute accent = spacing acute U+00B4 ISOdia
		txt = txt.replace('&AElig;', "AE") # latin capital letter AE = latin capital ligature AE U+00C6 ISOlat1
		txt = txt.replace('&aelig;', "ae") # latin small letter ae = latin small ligature ae U+00E6 ISOlat1
		txt = txt.replace('&Agrave;', "A") # latin capital letter A with grave = latin capital letter A grave U+00C0 ISOlat1
		txt = txt.replace('&agrave;', "a") # latin small letter a with grave = latin small letter a grave U+00E0 ISOlat1
		txt = txt.replace('&alefsym;', "%") # alef symbol = first transfinite cardinal U+2135 NEW
		txt = txt.replace('&Alpha;', "Alpha") # greek capital letter alpha U+0391
		txt = txt.replace('&alpha;', "alpha") # greek small letter alpha U+03B1 ISOgrk3
		txt = txt.replace('&amp;amp;','&')
		txt = txt.replace('&amp;;','&')
		txt = txt.replace('&amp;', "&") # ampersand U+0026 ISOnum		
		txt = txt.replace('&amp','&')		
		txt = txt.replace('&apos;', "'")
		txt = txt.replace('&Aring;', "A") # latin capital letter A with ring above = latin capital letter A ring U+00C5 ISOlat1
		txt = txt.replace('&aring;', "a") # latin small letter a with ring above = latin small letter a ring U+00E5 ISOlat1
		txt = txt.replace('&asymp;', "~~") # almost equal to = asymptotic to U+2248 ISOamsr
		txt = txt.replace('&Atilde;', "A") # latin capital letter A with tilde U+00C3 ISOlat1
		txt = txt.replace('&atilde;', "a") # latin small letter a with tilde U+00E3 ISOlat1
		txt = txt.replace('&auml;', "a")
		txt = txt.replace('&Auml;', "A")
		txt = txt.replace('&Auml;', "A") # latin capital letter A with diaeresis U+00C4 ISOlat1
		txt = txt.replace('&auml;', "a") # latin small letter a with diaeresis U+00E4 ISOlat1
		txt = txt.replace('&bdquo;', '"') # double low-9 quotation mark U+201E NEW
		txt = txt.replace('&Beta;', "Beta") # greek capital letter beta U+0392
		txt = txt.replace('&beta;', "beta") # greek small letter beta U+03B2 ISOgrk3
		txt = txt.replace('&brvbar;', "|") # broken bar = broken vertical bar U+00A6 ISOnum
		txt = txt.replace('&bull;', "o") # bullet = black small circle U+2022 ISOpub
		txt = txt.replace('&Ccedil;', "C") # latin capital letter C with cedilla U+00C7 ISOlat1
		txt = txt.replace('&ccedil;', "c") # latin small letter c with cedilla U+00E7 ISOlat1
		txt = txt.replace('&cedil;', " ") # cedilla = spacing cedilla U+00B8 ISOdia
		txt = txt.replace('&cent;', "c") # cent sign U+00A2 ISOnum
		txt = txt.replace('&Chi;', "Chi") # greek capital letter chi U+03A7
		txt = txt.replace('&chi;', "chi") # greek small letter chi U+03C7 ISOgrk3
		txt = txt.replace('&circ;', "") # modifier letter circumflex accent U+02C6 ISOpub
		txt = txt.replace('&cong;', "~=") # approximately equal to U+2245 ISOtech
		txt = txt.replace('&copy;', "(c)") # copyright sign U+00A9 ISOnum
		txt = txt.replace('&crarr;', "<-'") # downwards arrow with corner leftwards = carriage return U+21B5 NEW
		txt = txt.replace('&curren;', "$") # currency sign U+00A4 ISOnum
		txt = txt.replace('&dagger;', "|") # dagger U+2020 ISOpub
		txt = txt.replace('&Dagger;', "||") # double dagger U+2021 ISOpub
		txt = txt.replace('&darr;', "v") # downwards arrow U+2193 ISOnum
		txt = txt.replace('&dArr;', "v") # downwards double arrow U+21D3 ISOamsa
		txt = txt.replace('&deg;', "o") # degree sign U+00B0 ISOnum
		txt = txt.replace('&Delta;', "Delta") # greek capital letter delta U+0394 ISOgrk3
		txt = txt.replace('&delta;', "delta") # greek small letter delta U+03B4 ISOgrk3
		txt = txt.replace('&divide;', "/") # division sign U+00F7 ISOnum
		txt = txt.replace('&Eacute;', "E") # latin capital letter E with acute U+00C9 ISOlat1
		txt = txt.replace('&eacute;', "e") # latin small letter e with acute U+00E9 ISOlat1
		txt = txt.replace('&Ecirc;', "E") # latin capital letter E with circumflex U+00CA ISOlat1
		txt = txt.replace('&ecirc;', "e") # latin small letter e with circumflex U+00EA ISOlat1
		txt = txt.replace('&Egrave;', "E") # latin capital letter E with grave U+00C8 ISOlat1
		txt = txt.replace('&egrave;', "e") # latin small letter e with grave U+00E8 ISOlat1
		txt = txt.replace('&emsp;', " ") # em space U+2003 ISOpub
		txt = txt.replace('&ensp;', " ") # en space U+2002 ISOpub
		txt = txt.replace('&Epsilon;', "Epsilon") # greek capital letter epsilon U+0395
		txt = txt.replace('&epsilon;', "epsilon") # greek small letter epsilon U+03B5 ISOgrk3
		txt = txt.replace('&equiv;', "==") # identical to U+2261 ISOtech
		txt = txt.replace('&Eta;', "Eta") # greek capital letter eta U+0397
		txt = txt.replace('&eta;', "eta") # greek small letter eta U+03B7 ISOgrk3
		txt = txt.replace('&ETH;', "D") # latin capital letter ETH U+00D0 ISOlat1
		txt = txt.replace('&eth;', "d") # latin small letter eth U+00F0 ISOlat1
		txt = txt.replace('&Euml;', "E") # latin capital letter E with diaeresis U+00CB ISOlat1
		txt = txt.replace('&euml;', "e") # latin small letter e with diaeresis U+00EB ISOlat1
		txt = txt.replace('&euro;', "EU$") # euro sign U+20AC NEW
		txt = txt.replace('&fnof;', "f") # latin small f with hook = function = florin U+0192 ISOtech
		txt = txt.replace('&frac12;', "1/2") # vulgar fraction one half = fraction one half U+00BD ISOnum
		txt = txt.replace('&frac14;', "1/4") # vulgar fraction one quarter = fraction one quarter U+00BC ISOnum
		txt = txt.replace('&frac34;', "3/4") # vulgar fraction three quarters = fraction three quarters U+00BE ISOnum
		txt = txt.replace('&frasl;', "/") # fraction slash U+2044 NEW
		txt = txt.replace('&Gamma;', "Gamma") # greek capital letter gamma U+0393 ISOgrk3
		txt = txt.replace('&gamma;', "gamma") # greek small letter gamma U+03B3 ISOgrk3
		txt = txt.replace('&ge;', ">=") # greater-than or equal to U+2265 ISOtech
		txt = txt.replace('&gt;', ">") # greater-than sign U+003E ISOnum
		txt = txt.replace('&hArr;', "<=>") # left right double arrow U+21D4 ISOamsa
		txt = txt.replace('&harr;', "<->") # left right arrow U+2194 ISOamsa
		txt = txt.replace('&hellip;', "...") # horizontal ellipsis = three dot leader U+2026 ISOpub
		txt = txt.replace('&Iacute;', "I") # latin capital letter I with acute U+00CD ISOlat1
		txt = txt.replace('&iacute;', "i") # latin small letter i with acute U+00ED ISOlat1
		txt = txt.replace('&Icirc;', "I") # latin capital letter I with circumflex U+00CE ISOlat1
		txt = txt.replace('&icirc;', "i") # latin small letter i with circumflex U+00EE ISOlat1
		txt = txt.replace('&iexcl;', "!") # inverted exclamation mark U+00A1 ISOnum
		txt = txt.replace('&Igrave;', "E") # latin capital letter I with grave U+00CC ISOlat1
		txt = txt.replace('&igrave;', "i") # latin small letter i with grave U+00EC ISOlat1
		txt = txt.replace('&image;', "|I") # blackletter capital I = imaginary part U+2111 ISOamso
		txt = txt.replace('&Iota;', "Iota") # greek capital letter iota U+0399
		txt = txt.replace('&iota;', "iota") # greek small letter iota U+03B9 ISOgrk3
		txt = txt.replace('&iquest;', "?") # inverted question mark = turned question mark U+00BF ISOnum
		txt = txt.replace('&Iuml;', "I") # latin capital letter I with diaeresis U+00CF ISOlat1
		txt = txt.replace('&iuml;', "i") # latin small letter i with diaeresis U+00EF ISOlat1
		txt = txt.replace('&Kappa;', "Kappa") # greek capital letter kappa U+039A
		txt = txt.replace('&kappa;', "kappa") # greek small letter kappa U+03BA ISOgrk3
		txt = txt.replace('&Lambda;', "Lambda") # greek capital letter lambda U+039B ISOgrk3
		txt = txt.replace('&lambda;', "lambda") # greek small letter lambda U+03BB ISOgrk3
		txt = txt.replace('&laquo;', '"') # left-pointing double angle quotation mark = left pointing guillemet U+00AB ISOnum
		txt = txt.replace('&larr;', "<-") # leftwards arrow U+2190 ISOnum
		txt = txt.replace('&lArr;', "<=") # leftwards double arrow U+21D0 ISOtech
		txt = txt.replace('&ldquo;', '"') # left double quotation mark U+201C ISOnum
		txt = txt.replace('&le;', "<=") # less-than or equal to U+2264 ISOtech
		txt = txt.replace('&lrm;', "") # left-to-right mark U+200E NEW RFC 2070
		txt = txt.replace('&lsaquo;', '"') # single left-pointing angle quotation mark U+2039 ISO proposed
		txt = txt.replace('&lsquo;', '"') # left single quotation mark U+2018 ISOnum
		txt = txt.replace('&lt;', "<") # less-than sign U+003C ISOnum
		txt = txt.replace('&macr;', "-") # macron = spacing macron = overline = APL overbar U+00AF ISOdia
		txt = txt.replace('&mdash;', "---") # em dash U+2014 ISOpub
		txt = txt.replace('&micro;', "u") # micro sign U+00B5 ISOnum
		txt = txt.replace('&middot;', ".") # middle dot = Georgian comma = Greek middle dot U+00B7 ISOnum
		txt = txt.replace('&Mu;', "Mu") # greek capital letter mu U+039C
		txt = txt.replace('&mu;', "mu") # greek small letter mu U+03BC ISOgrk3
		txt = txt.replace('&nbsp;&nbsp;',' ')
		txt = txt.replace('&nbsp;', " ") # no-break space = non-breaking space U+00A0 ISOnum
		txt = txt.replace('&ndash;', "--") # en dash U+2013 ISOpub
		txt = txt.replace('&ne;', "!=") # not equal to U+2260 ISOtech
		txt = txt.replace('&not;', "-.") # not sign U+00AC ISOnum
		txt = txt.replace('&Ntilde;', "N") # latin capital letter N with tilde U+00D1 ISOlat1
		txt = txt.replace('&ntilde;', "n") # latin small letter n with tilde U+00F1 ISOlat1
		txt = txt.replace('&Nu;', "Nu") # greek capital letter nu U+039D
		txt = txt.replace('&nu;', "nu") # greek small letter nu U+03BD ISOgrk3
		txt = txt.replace('&Oacute;', "O") # latin capital letter O with acute U+00D3 ISOlat1
		txt = txt.replace('&oacute;', "o") # latin small letter o with acute U+00F3 ISOlat1
		txt = txt.replace('&Ocirc;', "O") # latin capital letter O with circumflex U+00D4 ISOlat1
		txt = txt.replace('&ocirc;', "o") # latin small letter o with circumflex U+00F4 ISOlat1
		txt = txt.replace('&OElig;', "OE") # latin capital ligature OE U+0152 ISOlat2
		txt = txt.replace('&oelig;', "oe") # latin small ligature oe U+0153 ISOlat2
		txt = txt.replace('&Ograve;', "O") # latin capital letter O with grave U+00D2 ISOlat1
		txt = txt.replace('&ograve;', "o") # latin small letter o with grave U+00F2 ISOlat1
		txt = txt.replace('&oline;', "-") # overline = spacing overscore U+203E NEW
		txt = txt.replace('&Omega;', "Omega") # greek capital letter omega U+03A9 ISOgrk3
		txt = txt.replace('&omega;', "omega") # greek small letter omega U+03C9 ISOgrk3
		txt = txt.replace('&Omicron;', "Omicron") # greek capital letter omicron U+039F
		txt = txt.replace('&omicron;', "omicron") # greek small letter omicron U+03BF NEW
		txt = txt.replace('&ordf;', "e") # feminine ordinal indicator U+00AA ISOnum
		txt = txt.replace('&ordm;', "o") # masculine ordinal indicator U+00BA ISOnum
		txt = txt.replace('&Oslash;', "O") # latin capital letter O with stroke = latin capital letter O slash U+00D8 ISOlat1
		txt = txt.replace('&oslash;', "o") # latin small letter o with stroke = latin small letter o slash U+00F8 ISOlat1
		txt = txt.replace('&Otilde;', "O") # latin capital letter O with tilde U+00D5 ISOlat1
		txt = txt.replace('&otilde;', "o") # latin small letter o with tilde U+00F5 ISOlat1
		txt = txt.replace('&Ouml;', "O") # latin capital letter O with diaeresis U+00D6 ISOlat1
		txt = txt.replace('&ouml;', "o") # latin small letter o with diaeresis U+00F6 ISOlat1
		txt = txt.replace('&para;', "|p") # pilcrow sign = paragraph sign U+00B6 ISOnum
		txt = txt.replace('&permil;', "%0") # per mille sign U+2030 ISOtech
		txt = txt.replace('&Phi;', "Phi") # greek capital letter phi U+03A6 ISOgrk3
		txt = txt.replace('&phi;', "phi") # greek small letter phi U+03C6 ISOgrk3
		txt = txt.replace('&Pi;', "Pi") # greek capital letter pi U+03A0 ISOgrk3
		txt = txt.replace('&pi;', "pi") # greek small letter pi U+03C0 ISOgrk3
		txt = txt.replace('&piv;', "pi") # greek pi symbol U+03D6 ISOgrk3
		txt = txt.replace('&plusmn;', "+-") # plus-minus sign = plus-or-minus sign U+00B1 ISOnum
		txt = txt.replace('&pound;', "p") # pound sign U+00A3 ISOnum
		txt = txt.replace('&Prime;', "''") # double prime = seconds = inches U+2033 ISOtech
		txt = txt.replace('&prime;', "'") # prime = minutes = feet U+2032 ISOtech
		txt = txt.replace('&Psi;', "Psi") # greek capital letter psi U+03A8 ISOgrk3
		txt = txt.replace('&psi;', "psi") # greek small letter psi U+03C8 ISOgrk3
		txt = txt.replace('&quot;', '"') # quotation mark = APL quote U+0022 ISOnum
		txt = txt.replace('&raquo;', '"') # right-pointing double angle quotation mark = right pointing guillemet U+00BB ISOnum
		txt = txt.replace('&rArr;', "=>") # rightwards double arrow U+21D2 ISOtech
		txt = txt.replace('&rarr;', "->") # rightwards arrow U+2192 ISOnum
		txt = txt.replace('&rdquo;', '"') # right double quotation mark U+201D ISOnum
		txt = txt.replace('&real;', "|R") # blackletter capital R = real part symbol U+211C ISOamso
		txt = txt.replace('&reg;', "(R)") # registered sign = registered trade mark sign U+00AE ISOnum
		txt = txt.replace('&Rho;', "Rho") # greek capital letter rho U+03A1
		txt = txt.replace('&rho;', "rho") # greek small letter rho U+03C1 ISOgrk3
		txt = txt.replace('&rlm;', "") # right-to-left mark U+200F NEW RFC 2070
		txt = txt.replace('&rsaquo;', '"') # single right-pointing angle quotation mark U+203A ISO proposed
		txt = txt.replace('&rsquo;', '"') # right single quotation mark U+2019 ISOnum
		txt = txt.replace('&sbquo;', '"') # single low-9 quotation mark U+201A NEW
		txt = txt.replace('&Scaron;', "S") # latin capital letter S with caron U+0160 ISOlat2
		txt = txt.replace('&scaron;', "s") # latin small letter s with caron U+0161 ISOlat2
		txt = txt.replace('&sdot;', ".") # dot operator U+22C5 ISOamsb
		txt = txt.replace('&sect;', "S") # section sign U+00A7 ISOnum
		txt = txt.replace('&shy;', "-") # soft hyphen = discretionary hyphen U+00AD ISOnum
		txt = txt.replace('&Sigma;', "Sigma") # greek capital letter sigma U+03A3 ISOgrk3
		txt = txt.replace('&sigma;', "sigma") # greek small letter sigma U+03C3 ISOgrk3
		txt = txt.replace('&sigmaf;', "sigma") # greek small letter final sigma U+03C2 ISOgrk3
		txt = txt.replace('&sim;', "~") # tilde operator = varies with = similar to U+223C ISOtech
		txt = txt.replace('&sup1;', "1") # superscript one = superscript digit one U+00B9 ISOnum
		txt = txt.replace('&sup2;', "2") # superscript two = superscript digit two = squared U+00B2 ISOnum
		txt = txt.replace('&sup3;', "3") # superscript three = superscript digit three = cubed U+00B3 ISOnum
		txt = txt.replace('&szlig;', "SS") # latin small letter sharp s = ess-zed U+00DF ISOlat1
		txt = txt.replace('&Tau;', "Tau") # greek capital letter tau U+03A4
		txt = txt.replace('&tau;', "tau") # greek small letter tau U+03C4 ISOgrk3
		txt = txt.replace('&Theta;', "Theta") # greek capital letter theta U+0398 ISOgrk3
		txt = txt.replace('&theta;', "theta") # greek small letter theta U+03B8 ISOgrk3
		txt = txt.replace('&thetasym;', "theta") # greek small letter theta symbol U+03D1 NEW
		txt = txt.replace('&thinsp;', " ") # thin space U+2009 ISOpub
		txt = txt.replace('&THORN;', "D") # latin capital letter THORN U+00DE ISOlat1
		txt = txt.replace('&thorn;', "d") # latin small letter thorn U+00FE ISOlat1
		txt = txt.replace('&tilde;', "~") # small tilde U+02DC ISOdia
		txt = txt.replace('&times;', "x") # multiplication sign U+00D7 ISOnum
		txt = txt.replace('&trade;', "tm") # trade mark sign U+2122 ISOnum
		txt = txt.replace('&Uacute;', "U") # latin capital letter U with acute U+00DA ISOlat1
		txt = txt.replace('&uacute;', "u") # latin small letter u with acute U+00FA ISOlat1
		txt = txt.replace('&uarr;', "^") # upwards arrow U+2191 ISOnum
		txt = txt.replace('&uArr;', "^") # upwards double arrow U+21D1 ISOamsa
		txt = txt.replace('&Ucirc;', "U") # latin capital letter U with circumflex U+00DB ISOlat1
		txt = txt.replace('&ucirc;', "u") # latin small letter u with circumflex U+00FB ISOlat1
		txt = txt.replace('&Ugrave;', "U") # latin capital letter U with grave U+00D9 ISOlat1
		txt = txt.replace('&ugrave;', "u") # latin small letter u with grave U+00F9 ISOlat1
		txt = txt.replace('&uml;', "''") # diaeresis = spacing diaeresis U+00A8 ISOdia
		txt = txt.replace('&upsih;', "upsilon") # greek upsilon with hook symbol U+03D2 NEW
		txt = txt.replace('&Upsilon;', "Upsilon") # greek capital letter upsilon U+03A5 ISOgrk3
		txt = txt.replace('&upsilon;', "upsilon") # greek small letter upsilon U+03C5 ISOgrk3
		txt = txt.replace('&Uuml;', "U") # latin capital letter U with diaeresis U+00DC ISOlat1
		txt = txt.replace('&uuml;', "u") # latin small letter u with diaeresis U+00FC ISOlat1
		txt = txt.replace('&weierp;', "P") # script capital P = power set = Weierstrass p U+2118 ISOamso
		txt = txt.replace('&Xi;', "Xi") # greek capital letter xi U+039E ISOgrk3
		txt = txt.replace('&xi;', "xi") # greek small letter xi U+03BE ISOgrk3
		txt = txt.replace('&Yacute;', "Y") # latin capital letter Y with acute U+00DD ISOlat1
		txt = txt.replace('&yacute;', "y") # latin small letter y with acute U+00FD ISOlat1
		txt = txt.replace('&yen;', "y") # yen sign = yuan sign U+00A5 ISOnum
		txt = txt.replace('&Yuml;', "Y") # latin capital letter Y with diaeresis U+0178 ISOlat2
		txt = txt.replace('&yuml;', "y") # latin small letter y with diaeresis U+00FF ISOlat1
		txt = txt.replace('&Zeta;', "Zeta") # greek capital letter zeta U+0396
		txt = txt.replace('&zeta;', "zeta") # greek small letter zeta U+03B6 ISOgrk3
		txt = txt.replace('&zwj;', "|") # zero width joiner U+200D NEW RFC 2070
		txt = txt.replace('&zwnj;', " ") # zero width non-joiner U+200C NEW RFC 2070

	return txt

#################################################################################################################
def isFloat(s):
	try:
		float(s)
		return True
	except:
		return False

#################################################################################################################
def isInt(s):
	try:
		int(s)
		return True
	except:
		return False

#################################################################################################################
# if success: returns the html page given in url as a string
# else: return -1 for Exception None for HTTP timeout, '' for empty page otherwise page data
#################################################################################################################
def fetchURL(url, file='', params='', cookies = [], timeout = 40, silent=False, isImage=False):
	safe_url = url
	debug("> fetchURL() url: " + safe_url)
	deleteTemp = (file == '')

	def SubthreadProc(url, params, cookies, result):
		try:
			if cookies or params:
				agent = urllib.URLopener()
				for cook in cookies:
					name, value = cook
					agent.addheader(name, value)

				f = agent.retrieve(url, file, data=params)
			else:
				f = urllib.urlretrieve(url, file)

		except IOError, errobj:
			result.append(-1)
			try:
				code = errobj[1]
			except:
				code = 0
			title, txt = HTTPErrorCode(code)
			result.append(title)
			result.append(txt)
		except:
			result.append(-1)
			result.append("Unknown exception")
		else:
			result.append(f)

	if file == '':
		dir = os.getcwd()
		if dir[-1]==';': dir=dir[0:-1]
		if dir[-1]!='\\': dir=dir+'\\'
		file = dir + "temp.dat"
	debug('file: ' + file)

	# remove destination file if exists already
	deleteFile(file)

	# call subthread to do fetch, then wait for result
	data = None
	title = ""
	msg1 = ""
	msg2 = ""
	result = []
	subThread = Thread(target=SubthreadProc, args=(safe_url, params, cookies, result))
	subThread.setDaemon(True)
	subThread.start()
	subThread.join(timeout)

	# check result, if OK, read data in from file
	if result == []:
		title = "HTTP Timeout"
		msg1 = "Try later."
	elif result[0] == -1:
		title = "HTTP Error"
		for msg in result[1:]:
			msg1 += msg + "\n"
	else:
		if not isImage:
			data = readFile(result[0][0])
			if not data:
				debug( "data file is empty")
				data = ''
		else:
			data = fileExist(file)
		urllib.urlcleanup()

	# report errors if not silent
	if not silent and (msg1 or msg2):
		messageOK(title, msg1, msg2)

	if deleteTemp:
		deleteFile(file)
	success = (data and data != -1)
	debug("< fetchURL() success = " + str(success))
	return data

#################################################################################################################
# extract URL and name from data
def extractURLName(data):
	# URL and name
	link = ''
	name = ''
	matches = re.search("href=[\"'](.*?)[\"'](?:.*?)>(.*?)<", data, re.IGNORECASE)
	if matches:
		link = matches.group(1)
		name = matches.group(2)
	else:
		name = cleanHTML(data)
	return link, name

#################################################################################################################
def searchRegEx(data, regex, flags=re.IGNORECASE):
	debug("> searchRegEx() regex: " + regex)
	matches = re.search(regex, data, flags)
	if matches:
		value = matches.group(1)
	else:
		value = ""
	debug ("< searchRegEx() return: " + value)
	return value

#################################################################################################################
def findAllRegEx(data, regex, flags=re.MULTILINE+re.IGNORECASE):
	debug("> findAllRegEx() regex: " + regex)

	findRe = re.compile(regex, flags)
	matches = findRe.findall(data)

	if matches:
		sz = len(matches)
	else:
		sz = 0
	debug ("< findAllRegEx() sz: " + str(sz))
	return matches

#################################################################################################################
# Does a direct image URL exist in any RSS object elements?
def isImageURL(url):
	match = re.search('([^"\'<>]+)(?<=(?:\.gif|\.jpg|\.bmp))', url, re.IGNORECASE)
	try:
		url = match.group(1)
	except:
		url = ''
	debug("isImageURL() found: " + url)
	return url

#################################################################################################################
def isRSSLink(url):
	debug("isRSSLink()")
	return searchRegEx(url, '(rss|rdf|xml)$')

#################################################################################################################
def isHTMLLink(url):
	debug("isHTMLLink()")
	return searchRegEx(url, '(html|htm)$')

#################################################################################################################
class TextBoxDialog(xbmcgui.WindowDialog):
	def __init__(self):
		if Emulating: xbmcgui.WindowDialog.__init__(self)
		setResolution(self)

	def setupDisplay(self, isMC360):
		debug("> setupDisplay()")

		dialogW = REZ_W - 100
		dialogH = REZ_H - 100
		xpos = int((REZ_W /2) - (dialogW /2))
		ypos = int((REZ_H /2) - (dialogH /2)) +10

		if isMC360:
			DIALOG_PANEL = DIR_GFX + PANEL_FILENAME_MC360
		else:
			DIALOG_PANEL = DIR_GFX + PANEL_FILENAME

		try:
			self.addControl(xbmcgui.ControlImage(xpos, ypos, dialogW, dialogH, DIALOG_PANEL))
		except: pass

		xpos += 20
		ypos += 20
		self.titleCL = xbmcgui.ControlLabel(xpos, ypos, dialogW-35, 30, '','font14', "0xFFFFFF00", alignment=XBFONT_CENTER_X)
		self.addControl(self.titleCL)
		ypos += 30
		self.descCTB = xbmcgui.ControlTextBox(xpos, ypos, dialogW-35, dialogH-70, 'font12', '0xFFFFFF66')
		self.addControl(self.descCTB)
		debug("< setupDisplay()")


	def ask(self, file="", title="", text="", isMC360=False):
		debug("> TextBoxDialog().ask()")
		if not title and file:
			head, title = os.path.split(file)
		self.setupDisplay(isMC360)
		self.titleCL.setLabel(title)
		txt = text
		if text:
			txt += "\n\n"
		if file:
			txt += readFile(file)
		self.descCTB.setText(txt)
		self.setFocus(self.descCTB)
		self.doModal()
		debug("< TextBoxDialog().ask()")

	def onAction(self, action):
		debug("onAction()")
		if action == ACTION_PREVIOUS_MENU or action == ACTION_PARENT_DIR:
			self.close()

	def onControl(self, control):
		debug("onControl()")

#################################################################################################################
def listDir(path, ext='', fnRE='', getFullFilename=False, lower=False, upper=False):
	debug("> listDir()")

	fileList = []
	if ext:
		ext = ext.lower()
		if ext[0] != '.':
			ext = '.'+ext

	if os.path.isdir(path):
		files = os.listdir(path)
		for f in files:
			fn, ex = os.path.splitext(f)
			if not ext or ex.lower() == ext:
				found = False
				if fnRE:
					matches = re.search(fnRE, fn)
					if matches:
						found = True
				else:
					found = True

				if found:
					if not getFullFilename:
						filename = fn
					else:
						filename = f

					if lower:
						fileList.append(filename.lower())
					elif upper:
						fileList.append(filename.upper())
					else:
						fileList.append(filename)

	sz = len(fileList)
	debug("< listDir() file count="+str(sz))
	return fileList