############################################################################################################
# TV Data source: US_SageTV based on Zap2it
#
# SETUP
# -----
#
# 25/07/2007 First attempt
############################################################################################################

from mytvLib import *
import xbmc,xbmcgui, re, time
from string import split
import time

# global vars
dialog = xbmcgui.DialogProgress()
# DO NOT CHANGE THESE - USERNAME & PASSWORD SETUP IS NOW DONE ON FIRST RUN OR THROU CONFIG MENU
CONFIG_OPT_IP = 'IP'
CONFIG_OPT_PORT = 'port'
CONFIG_OPT_USER = 'username'
CONFIG_OPT_PASS = 'password'

class ListingData:
	def __init__(self, cache):
		self.name = os.path.splitext(os.path.basename( __file__))[0]	# get filename without path & ext
		self.configSection = 'DATASOURCE_' + self.name
		self.config()

		self.BASE_URL = "http://" + self.username + ":" + self.password + "@" + self.IP + ":" + self.port + "/sage/"
		self.channelListURL = self.BASE_URL + "EpgGrid"	# all channels group
		self.channelURL = self.BASE_URL + "Search?SearchString=&searchType=Airings&Video=on&DVD=on&search_fields=title&TimeRange=120&Categories=**Any**&Channels=$CHID&watched=any&dontlike=any&favorite=any&autodelete=any&partials=none&sort1=airdate_asc&sort2=none&pagelen=inf"
		self.CHANNELS_FILENAME = cache + "Channels_"+ self.name + ".dat"
		self.DIR_CACHE = cache

	def getName(self):
		return self.name


	# download or load if exists, a list of all available channels.
	# return: list [chID, chName]
	def getChannels(self):
		url = self.channelListURL
		regex = 'value="(\d+)".*?>\d+ - (.*?)<'
		return getChannelsLIB(url, self.CHANNELS_FILENAME , \
							regex, startStr='startchan',endStr='</body>')


	# download channel data, using either dayDelta or dataDate.
	# filename = filename to save downloaded data file as.
	# chID = unique channel ID, used to reference channel in URL.
	# chName = display name of channel
	# dayDelta = day offset from today. ie 0 is today, 1 is tomorrow ...
	# fileDate = use to calc programme start time in secs since epoch.
	# return Channel class or -1 if http fetch error, or None for other
	def getChannel(self, filename, chID, chName, dayDelta, fileDate):
		debug("getChannel() dayDelta: " + str(dayDelta) + " chID: " + chID + " fileDate: " + fileDate)
		programmeList = []
		lastStartTimeSecs = 0 
		filename = ''
		if dayDelta < 0:
			return []	# no data as this site doesnt have negative day fetches

		# download data file if file doesnt exists
		if not fileExist(filename):
			formatTime = time.strptime(fileDate,"%Y%m%d")
			urlDate = time.strftime("%d-%m-%Y", formatTime)
			url = self.channelURL.replace('$CHID',chID)
			displayDate = time.strftime("%a %d %b",time.strptime(fileDate,"%Y%m%d"))
			dialog.create("Downloading ...","Channel: " + chName,"Date: " + displayDate)
			doc = fetchURL(url, filename).decode('utf8','replace')
			dialog.close()
		else:
			doc = readFile(filename).decode('utf8','replace')

		# check for timout, exception error, or empty page
		if not doc or doc == -1:
			return doc

		debug("regex title & link ...")

		# time, id, title, desc
		regex = 'AiringId=(\d+)">(.*?)(</a>|<br/>)(.*?)(</div>|</a>).*?(Markerblank|Manual|Favorite).*?datecell">.*?>(.*?),(.*?) (\d+).*?(\d+):(\d\d) (.M).*?-.*?(\d+):(\d\d) (.M)'
		matches = parseDocList(doc, '','',regex)
		month = {'Jan': '01', 'Feb': '02', 'Mar': '03', 'Apr': '04', 'May': '05', 'Jun': '06',
				'Jul': '07', 'Aug': '08', 'Sep': '09', 'Oct': '10', 'Nov': '11', 'Dec': '12'}

		seenAM = False	# ignore all times starting on each page, until we get to an AM
		for match in matches:
			current_day = '0' + match[7]
#			if fileDate[4:6] != current_day[-2:]
#				continue
			startTime = match[1]
			if match[9] == '12':
				if match[11] == 'PM':
					start_time = '12'
				else:
					start_time = '00'
			elif match[11] == 'PM':
				calc_time = int(match[9]) + 12
				start_time = str(calc_time)
			else:
				start_time = match[9]
			if len(start_time) < 2:
				start_time = '0' + start_time
			start_minute = match[10]
			startTime = start_time + ':' + start_minute
			id = match[0]
			title = cleanHTML(decodeEntities(match[1]))
			desc = ''
			if match[5] != "Markerblank":
				genre = match[5]
			else:
				genre = ""
			if not startTime or not title:
				continue

#			if not seenAM:
#				if int(startTime[:2]) >= 12:
#					continue
#				seenAM = True
#
			# convert starttime to secs since epoch
			startTimeSecs = startTimeToSecs(lastStartTimeSecs, startTime, fileDate)
			lastStartTimeSecs = startTimeSecs
#			print startTimeSecs, id, title
			programmeList.append(Programme(title, desc, startTimeSecs, 0, id, genre))

		# set end times to be that of next prog start time. apart from last prog set endtime = 0
		# save last prog details
		for progIDX in range(len(programmeList)):
			try:
				programmeList[progIDX].setEndTime(programmeList[progIDX+1].getStartTime())
			except:
				# last prog
				programmeList[progIDX].setEndTime(0)

		return programmeList


	############################################################################################################
	# Download url and regex parse it to extract description.
	############################################################################################################
	def getLink(self, link, title):
		debug("> getLink()")
		desc = ''
		html = ''

		det_URL = self.BASE_URL + "DetailedInfo?AiringId=" + link
		detail = fetchURL(det_URL, '').decode('utf8','replace')
		try:
			info = re.compile('<p>Description:(.*?)</p>', re.DOTALL).findall(detail)
			desc = info[0]
		except:
			desc = ' '

		return desc


	############################################################################################################
	def isUserSetup(self):
		if not self.username or not self.password:
			messageOK("Alert","Username/Password missing.","Please configure datasource.")
			return False
		else:
			return True

	############################################################################################################
	# load, if not exist ask, then save
	############################################################################################################
	def config(self, reset=False):
		debug("> config()")

		if reset:
			self.IP = ''
			self.port = ''
			self.username = ''
			self.password = ''

		# IP
		if not reset:
			self.IP = config.action(CONFIG_OPT_IP, section=self.configSection)
		while not self.IP:
			messageOK("Configure DataSource","Enter SageTV webserver IP")
			keyboard = xbmc.Keyboard()
			keyboard.doModal()
			if keyboard.isConfirmed:
				self.IP = keyboard.getText().strip()
				if self.IP:
					config.action(CONFIG_OPT_IP, self.IP, section=self.configSection, mode=Config.MODE_WRITE)
			else:
				break

		# PORT
		if not reset:
			self.port = config.action(CONFIG_OPT_PORT, section=self.configSection)
		while not self.port:
			messageOK("Configure DataSource","Please enter Port")
			keyboard = xbmc.Keyboard()
			keyboard.doModal()
			if keyboard.isConfirmed:
				self.port = keyboard.getText().strip()
				if self.port:
					config.action(CONFIG_OPT_PORT, self.port, section=self.configSection, mode=Config.MODE_WRITE)
			else:
				break


		# USERNAME
		if not reset:
			self.username = config.action(CONFIG_OPT_USER, section=self.configSection)
		while not self.username:
			messageOK("Configure DataSource","Please enter Username")
			keyboard = xbmc.Keyboard()
			keyboard.doModal()
			if keyboard.isConfirmed:
				self.username = keyboard.getText().strip()
				if self.username:
					config.action(CONFIG_OPT_USER, self.username, section=self.configSection, mode=Config.MODE_WRITE)
			else:
				break

		# PASSWORD
		if not reset:
			self.password = config.action(CONFIG_OPT_PASS, section=self.configSection)
		while not self.password:
			messageOK("Configure DataSource","Please enter Password")
			keyboard = xbmc.Keyboard()
			keyboard.doModal()
			if keyboard.isConfirmed:
				self.password = keyboard.getText().strip()
				if self.password:
					config.action(CONFIG_OPT_PASS, self.password, section=self.configSection, mode=Config.MODE_WRITE)
			else:
				break



		success = (self.username and self.password)
		debug("< config() success="+str(success))
		return success
